"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.stringifyValue = stringifyValue;
exports.replaceWithCode = replaceWithCode;
exports.default = callDump;

var _babelPluginMacros = require("babel-plugin-macros");

var _core = require("@babel/core");

var _macroAssertions = require("./macro-assertions");

var _javascriptStringify = require("javascript-stringify");

function stringifyValue(val, varName) {
  const stringified = (0, _javascriptStringify.stringify)(val);

  if (stringified === undefined) {
    (0, _macroAssertions.throwUnexpectedError)(`Failed to stringify ${varName}, with value: ${val}`);
  }

  return stringified;
}

function replaceWithCode(code, path) {
  const ast = (0, _core.parse)(code);

  if (_core.types.isFile(ast)) {
    path.replaceWith(ast.program.body[0]);
  } else {
    (0, _macroAssertions.throwUnexpectedError)(`${code} was incorrectly parsed. The AST was: ${JSON.stringify(ast)}`);
  }
}

function dumpValue(path, types, exportedName, copyAll = false) {
  const typeNames = copyAll ? types.keys() : (0, _macroAssertions.getStringParameters)(path, exportedName);
  const selectedTypes = new Map();

  for (const name of typeNames) {
    const type = types.get(name);

    if (type === undefined) {
      throw new _babelPluginMacros.MacroError(`Failed to find type "${name}" in namedTypes`);
    }

    selectedTypes.set(name, type);
  }

  const stringified = stringifyValue(selectedTypes, "selectedTypes");
  replaceWithCode(stringified, copyAll ? path : path.parentPath);
}

function dumpValues(paths, namedTypes, exportedName, copyAll = false) {
  for (const path of paths) {
    dumpValue(path, namedTypes, exportedName, copyAll);
  }
}

function callDump(references, namedTypes, dumpName, copyAll = false) {
  const paths = references[dumpName];

  if (paths !== undefined) {
    dumpValues(paths, namedTypes, dumpName, copyAll);
    return true;
  }

  return false;
}