"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.assertArray = assertArray;
exports.assertPrimitiveType = assertPrimitiveType;
exports.assertPrimitiveTypeName = assertPrimitiveTypeName;
exports.assertLiteral = assertLiteral;
exports.assertAcceptsTypeParameters = assertAcceptsTypeParameters;
exports.assertObjectPattern = assertObjectPattern;
exports.assertBuiltinType = assertBuiltinType;
exports.Union = Union;
exports.Tuple = Tuple;
exports.Intersection = Intersection;
exports.Literal = Literal;
exports.PrimitiveType = PrimitiveType;
exports.GenericType = GenericType;
exports.Type = Type;
exports.BuiltinType = BuiltinType;
exports.FailedIntersection = FailedIntersection;
exports.PropertySignature = PropertySignature;
exports.isArrayType = exports.isIntersectionOrUnion = exports.isAnyOrUnknown = exports.isObjectPattern = exports.isFailedIntersection = exports.isLiteral = exports.isTuple = exports.isInstantiatedType = exports.isBuiltinType = exports.isIntersection = exports.isUnion = exports.isGenericType = exports.isTypeAlias = exports.isInterface = exports.isPrimitive = exports.isType = void 0;

var _babelPluginMacros = require("babel-plugin-macros");

var _macroAssertions = require("../macro-assertions");

var _checks = require("../utils/checks");

var _IR = require("./IR");

// These MUST be kept in sync with IR.ts
const isType = x => x.type === "type";

exports.isType = isType;

const isPrimitive = x => x.type === "primitiveType";

exports.isPrimitive = isPrimitive;

const isInterface = x => x.type === "interface";

exports.isInterface = isInterface;

const isTypeAlias = x => x.type === "alias";

exports.isTypeAlias = isTypeAlias;

const isGenericType = x => x.type === "genericType";

exports.isGenericType = isGenericType;

const isUnion = x => x.type === "union";

exports.isUnion = isUnion;

const isIntersection = x => x.type === "intersection"; // Is this sound/enough?


exports.isIntersection = isIntersection;

const isBuiltinType = x => x.type === "builtinType";

exports.isBuiltinType = isBuiltinType;

const isInstantiatedType = x => x.type === "instantiatedType";

exports.isInstantiatedType = isInstantiatedType;

const isTuple = x => x.type === "tuple";

exports.isTuple = isTuple;

const isLiteral = x => x.type === "literal";

exports.isLiteral = isLiteral;

const isFailedIntersection = x => x.type === "failedIntersection";

exports.isFailedIntersection = isFailedIntersection;

const isObjectPattern = x => x.type === "objectPattern";

exports.isObjectPattern = isObjectPattern;

const isAnyOrUnknown = x => {
  return isPrimitive(x) && (x.typeName === "any" || x.typeName === "unknown");
};

exports.isAnyOrUnknown = isAnyOrUnknown;

const isIntersectionOrUnion = x => isIntersection(x) || isUnion(x);

exports.isIntersectionOrUnion = isIntersectionOrUnion;

const isArrayType = x => isType(x) && (x.typeName === "Array" || x.typeName === "ReadonlyArray");

exports.isArrayType = isArrayType;

function assertArray(x) {
  if (!isArrayType(x)) (0, _macroAssertions.throwUnexpectedError)(`expected array but recieved: ${x.type}`);
}

function assertPrimitiveType(x) {
  if (!isPrimitive(x)) {
    (0, _macroAssertions.throwUnexpectedError)(`expected ${JSON.stringify(x)} to be a primitive type`);
  }
}

function assertPrimitiveTypeName(type) {
  if (!_IR.primitiveTypes.includes(type)) {
    (0, _macroAssertions.throwUnexpectedError)(`${type} is not a primitive type`);
  }
}

function assertLiteral(x) {
  if (!isLiteral(x)) {
    (0, _macroAssertions.throwUnexpectedError)(`Expected ${JSON.stringify(x)} to be a literal type`);
  }
}

function assertAcceptsTypeParameters(ir, typeName) {
  if (!isInterface(ir) && !isTypeAlias(ir) && !isBuiltinType(ir)) {
    throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.TypeDoesNotAcceptGenericParameters(typeName, ir.type));
  }
}

function assertObjectPattern(ir) {
  if (!isObjectPattern(ir)) {
    (0, _macroAssertions.throwUnexpectedError)(`Expected ${JSON.stringify(ir)} to be an object pattern`);
  }
}

function assertBuiltinType(ir, typeName) {
  if (isBuiltinType(ir)) {
    if (ir.typeName === typeName) return;
  }

  (0, _macroAssertions.throwUnexpectedError)(`Expected ${JSON.stringify(ir)} to be a builtin ${typeName}`);
}

function Union(hasUndefined, ...childTypes) {
  const union = {
    type: "union",
    childTypes,
    ...(hasUndefined && {
      hasUndefined
    })
  };
  return union;
}

function Tuple({
  childTypes,
  restType,
  firstOptionalIndex,
  undefinedOptionals = true
}) {
  const tuple = {
    type: "tuple",
    childTypes,
    restType,
    firstOptionalIndex,
    undefinedOptionals
  };
  return tuple;
}

function Intersection(...childTypes) {
  const intersection = {
    type: "intersection",
    childTypes
  };
  return intersection;
}

function Literal(value) {
  const literal = {
    type: "literal",
    value
  };
  return literal;
}

function PrimitiveType(typeName) {
  assertPrimitiveTypeName(typeName);
  const primitive = {
    type: "primitiveType",
    typeName
  };
  return primitive;
}

function GenericType(typeParameterIndex) {
  const genericType = {
    type: "genericType",
    typeParameterIndex
  };
  return genericType;
}

function Type(typeName, ...typeParameters) {
  if ((0, _checks.hasAtLeast1Element)(typeParameters)) {
    const type = {
      type: "type",
      typeName,
      typeParameters
    };
    return type;
  }

  const type = {
    type: "type",
    typeName
  };
  return type;
}

function BuiltinType(typeName, elementType1, elementType2) {
  if (typeName === "Map") {
    const builtinType = {
      type: "builtinType",
      typeName: "Map",
      elementTypes: [elementType1, elementType2],
      typeParametersLength: 2,
      typeParameterDefaults: []
    };
    return builtinType;
  } else {
    const builtinType = {
      type: "builtinType",
      typeName: typeName,
      elementTypes: [elementType1],
      typeParametersLength: 1,
      typeParameterDefaults: []
    };
    return builtinType;
  }
}

function FailedIntersection() {
  return {
    type: "failedIntersection"
  };
}

function PropertySignature(keyName, optional, value) {
  return {
    type: "propertySignature",
    keyName,
    optional,
    value
  };
}