"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _utils = require("./utils");

var u = _interopRequireWildcard(require("../IRUtils"));

var _macroAssertions = require("../../macro-assertions");

var _commonTags = require("common-tags");

var _checks = require("../../utils/checks");

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function cleanUnion(union, instantiatedTypes) {
  let childTypes = union.childTypes;
  childTypes = childTypes.filter(t => !(0, u.isFailedIntersection)(t));

  if (childTypes.length === 0) {
    // TODO: Convert this to formal error
    (0, _macroAssertions.throwMaybeAstError)((0, _commonTags.oneLine)`union had 0 child types after stripping failed intersections. 
        You probably have an invalid intersection in your code.`);
  }

  const primitives = new Set();

  function hasSubsumingPrimitive(x) {
    if (u.isInstantiatedType(x)) {
      x = (0, _utils.getInstantiatedType)(x.typeName, instantiatedTypes);
    }

    if (u.isLiteral(x) && primitives.has(typeof x.value)) {
      return true;
    } else if (u.isObjectPattern(x) && primitives.has("object")) {
      return true;
    }

    return false;
  }

  const notPrimitives = [];

  for (let type of childTypes) {
    if (u.isInstantiatedType(type)) {
      const instantiated = (0, _utils.getInstantiatedType)(type.typeName, instantiatedTypes);

      if (u.isPrimitive(instantiated)) {
        type = instantiated;
      }

      if (u.isUnion(instantiated)) {
        (0, _macroAssertions.throwUnexpectedError)(`there should be any nested unions in the cleaning pass: ${JSON.stringify(childTypes)}`);
      }
    }

    if ((0, u.isAnyOrUnknown)(type)) {
      return u.PrimitiveType("any");
    }

    if (u.isPrimitive(type)) {
      primitives.add(type.typeName);
    } else {
      notPrimitives.push(type);
    }
  }

  childTypes = notPrimitives.filter(t => !hasSubsumingPrimitive(t));

  for (const typeName of primitives) {
    childTypes.push(u.PrimitiveType(typeName));
  }

  if (childTypes.length === 1) {
    return childTypes[0];
  }

  (0, _checks.hasAtLeast2Elements)(childTypes);
  return u.Union(primitives.has("undefined"), childTypes[0], childTypes[1], ...childTypes.slice(2));
}

function cleanUnions(ir, instantiatedTypes) {
  return (0, _utils.traverse)(ir, u.isUnion, cur => {
    return cleanUnion(cur, instantiatedTypes);
  });
}

var _default = cleanUnions;
exports.default = _default;