"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;

var _babelPluginMacros = require("babel-plugin-macros");

var _macroAssertions = require("../../macro-assertions");

var _utils = require("./utils");

var _IRUtils = require("../IRUtils");

var _fastCopy = _interopRequireDefault(require("fast-copy"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// We can collect type stats before merging unions / solving intersections
// because types with identical names in unions will be merged in the flatten stage
function addStats(base, toAdd) {
  for (const [k, v] of toAdd) {
    if (base.has(k)) {
      base.set(k, toAdd.get(k) + v);
    } else {
      base.set(k, v);
    }
  }
}

function incrementTypeCount(typeInfo, typeName) {
  if (!typeInfo.has(typeName)) {
    typeInfo.set(typeName, 1);
  } else {
    typeInfo.set(typeName, typeInfo.get(typeName) + 1);
  }
}

function _default(ir, state) {
  const circularTypes = [];
  const patchedIr = patchIR(ir, { ...state,
    circularTypes
  }, new Set());

  for (const type of circularTypes) {
    const typeInfo = state.instantiatedTypes.get(type);

    if (typeInfo === undefined) {
      (0, _macroAssertions.throwUnexpectedError)(`Type: ${type} was not found in instantiatedTypes`);
    }

    typeInfo.circular = true;
  }

  return patchedIr;
}

function patchIR(ir, state, callerTypeNames) {
  const {
    namedTypes,
    typeStats,
    instantiatedTypes,
    newInstantiatedTypes
  } = state;
  return (0, _utils.traverse)(ir, _IRUtils.isType, typeRef => {
    const {
      typeName,
      typeParameters: providedTypeParameters = []
    } = typeRef;
    const key = (0, _utils.getTypeKey)(typeRef);
    const partiallyResolvedTypeReference = {
      type: "instantiatedType",
      typeName: key
    };
    incrementTypeCount(typeStats, key);

    if (callerTypeNames.has(key)) {
      state.circularTypes.push(key);
      return partiallyResolvedTypeReference;
    }

    let partiallyResolved = instantiatedTypes.get(key);

    if (partiallyResolved !== undefined) {
      addStats(typeStats, partiallyResolved.typeStats);
      return partiallyResolvedTypeReference;
    }

    const referencedIR = namedTypes.get(typeName);

    if (referencedIR === undefined) {
      throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.UnregisteredType(typeName));
    }

    (0, _IRUtils.assertAcceptsTypeParameters)(referencedIR, typeName);
    const typeParametersApplied = (0, _utils.applyTypeParameters)(referencedIR, typeName, providedTypeParameters);
    const newState = { ...state,
      typeStats: new Map()
    };
    const newCallerTypeNames = (0, _fastCopy.default)(callerTypeNames);
    const instantiated = patchIR(typeParametersApplied, newState, newCallerTypeNames.add(key));
    instantiatedTypes.set(key, {
      typeStats: newState.typeStats,
      value: instantiated,
      circular: false
    });
    newInstantiatedTypes.push(key);
    addStats(typeStats, newState.typeStats);
    return partiallyResolvedTypeReference;
  });
}