"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = resolveAllTypes;
exports.resolveSingleType = resolveSingleType;

var _babelPluginMacros = require("babel-plugin-macros");

var _fastCopy = _interopRequireDefault(require("fast-copy"));

var _macroAssertions = require("../../macro-assertions");

var _IR = require("../IR");

var _IRUtils = require("../IRUtils");

var _utils = require("./utils");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function resolveAllTypes(namedTypes) {
  const cloned = (0, _fastCopy.default)(namedTypes);

  for (const [typeName, ir] of namedTypes) {
    if (_IR.builtinTypes.includes(typeName)) continue;
    namedTypes.set(typeName, resolveSingleType(ir, cloned, typeName));
  }
}

function resolveSingleType(ir, namedTypes, typeName) {
  return resolveType(ir, {
    namedTypes,
    visitedTypes: new Set([typeName])
  }, (0, _IRUtils.isTypeAlias)(ir) ? _IRUtils.isObjectPattern : undefined);
}

function resolveType(ir, state, skip) {
  const {
    namedTypes,
    visitedTypes
  } = state;
  return (0, _utils.traverse)(ir, _IRUtils.isType, typeRef => {
    const {
      typeName,
      typeParameters = []
    } = typeRef;
    if (visitedTypes.has(typeName)) return typeRef;
    const referencedIR = namedTypes.get(typeName); // TODO: Handle Array, Map, Record

    if (referencedIR === undefined) {
      throw new _babelPluginMacros.MacroError(_macroAssertions.Errors.UnregisteredType(typeName));
    }

    if ((0, _IRUtils.isTypeAlias)(referencedIR)) {
      const instantiatedIR = (0, _utils.applyTypeParameters)(referencedIR, typeName, typeParameters);
      const newVisited = (0, _fastCopy.default)(visitedTypes);
      newVisited.add(typeName);
      return resolveType(instantiatedIR, { ...state,
        visitedTypes: newVisited
      }, _IRUtils.isObjectPattern);
    } else if ((0, _IRUtils.isInterface)(referencedIR) || (0, _IRUtils.isBuiltinType)(referencedIR)) {
      return typeRef;
    } else {
      (0, _macroAssertions.throwUnexpectedError)(`type reference referenced a ${referencedIR.type} instead of an interface or alias`);
    }
  }, skip);
}