"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _babelPluginMacros = require("babel-plugin-macros");

var _fastCopy = _interopRequireDefault(require("fast-copy"));

var _irToInline = _interopRequireDefault(require("./code-gen/irToInline"));

var _dumpHelpers = _interopRequireWildcard(require("./dump-helpers"));

var _macroAssertions = require("./macro-assertions");

var _register = require("./register");

var _astToTypeIR = require("./type-ir/astToTypeIR");

var _IR = require("./type-ir/IR");

var u = _interopRequireWildcard(require("./type-ir/IRUtils"));

var _clean = _interopRequireDefault(require("./type-ir/passes/clean"));

var _flatten = _interopRequireDefault(require("./type-ir/passes/flatten"));

var _instantiate = _interopRequireDefault(require("./type-ir/passes/instantiate"));

var _intersect = _interopRequireDefault(require("./type-ir/passes/intersect"));

var _resolve = _interopRequireDefault(require("./type-ir/passes/resolve"));

function _getRequireWildcardCache() { if (typeof WeakMap !== "function") return null; var cache = new WeakMap(); _getRequireWildcardCache = function () { return cache; }; return cache; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } if (obj === null || typeof obj !== "object" && typeof obj !== "function") { return { default: obj }; } var cache = _getRequireWildcardCache(); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj.default = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const baseNamedTypes = new Map([["Array", u.BuiltinType("Array", u.GenericType(0), undefined)], ["Set", u.BuiltinType("Set", u.GenericType(0), undefined)], ["Map", u.BuiltinType("Map", u.GenericType(0), u.GenericType(1))]]);

function removeBuiltins(map) {
  const builtinsRemoved = (0, _fastCopy.default)(map);

  for (const builtin of _IR.builtinTypes) {
    builtinsRemoved.delete(builtin);
  }

  return builtinsRemoved;
}

function finalizeType(path, instantiatedTypes, namedTypes) {
  const typeParam = (0, _macroAssertions.getTypeParameter)(path);
  let ir = (0, _astToTypeIR.getTypeParameterIR)(typeParam.node);
  const state = {
    instantiatedTypes,
    namedTypes,
    typeStats: new Map(),
    newInstantiatedTypes: []
  }; // no type resolution on the type parameter

  ir = (0, _flatten.default)(ir);
  const instantiatedIR = (0, _instantiate.default)(ir, state);

  for (const type of state.newInstantiatedTypes) {
    const newType = instantiatedTypes.get(type);

    if (newType === undefined) {
      (0, _macroAssertions.throwUnexpectedError)(`did not expected ${type} to be undefined`);
    }

    newType.value = (0, _clean.default)((0, _intersect.default)(newType.value, instantiatedTypes), instantiatedTypes);
    instantiatedTypes.set(type, newType);
  }

  const finalIR = (0, _clean.default)((0, _intersect.default)(instantiatedIR, instantiatedTypes), instantiatedTypes);
  return [finalIR, state.typeStats];
} // @ts-ignore - @types/babel-plugin-macros is out of date


function macroHandler({
  references,
  state,
  babel
}) {
  const namedTypes = (0, _fastCopy.default)(baseNamedTypes);
  const instantiatedTypes = new Map();
  const registerExportName = "registerType";

  if (references[registerExportName]) {
    for (const path of references[registerExportName]) {
      const callExpr = path.parentPath;
      const typeName = (0, _macroAssertions.getRegisterArguments)(path);
      const stmtsInSameScope = (0, _macroAssertions.getBlockParent)(path);
      (0, _register.registerType)(typeName, stmtsInSameScope, namedTypes);
      callExpr.remove();
    }
  }

  if ((0, _dumpHelpers.default)(references, removeBuiltins(namedTypes), "__dumpAfterRegistration", true)) return;
  (0, _resolve.default)(namedTypes);
  if ((0, _dumpHelpers.default)(references, removeBuiltins(namedTypes), "__dumpAfterTypeResolution")) return;

  for (const [typeName, ir] of namedTypes) {
    if (_IR.builtinTypes.includes(typeName)) continue;
    namedTypes.set(typeName, (0, _flatten.default)(ir));
  }

  if ((0, _dumpHelpers.default)(references, removeBuiltins(namedTypes), "__dumpAfterTypeFlattening")) return;
  const dumpInstantiatedName = "__dumpInstantiatedIR";

  if (references[dumpInstantiatedName]) {
    for (const path of references[dumpInstantiatedName]) {
      const callExpr = path.parentPath;
      const instantiatedTypesToDump = new Map();
      const [finalIR, typeStats] = finalizeType(path, instantiatedTypesToDump, namedTypes);
      instantiatedTypesToDump.set("$$typeParameter$$", {
        typeStats,
        value: finalIR,
        circular: false
      });
      const stringified = (0, _dumpHelpers.stringifyValue)(instantiatedTypesToDump, "instantiatedTypes");
      (0, _dumpHelpers.replaceWithCode)(stringified, callExpr);
    }

    return;
  }

  if (references.default) {
    for (const path of references.default) {
      const callExpr = path.parentPath;
      const [finalIR, typeStats] = finalizeType(path, instantiatedTypes, namedTypes);
      const {
        circularRefs
      } = (0, _macroAssertions.getOptions)("boolean", path, "typecheck.macro's default export");
      const code = (0, _irToInline.default)(finalIR, {
        instantiatedTypes,
        options: {
          errorMessages: false,
          expectedValueAsIR: false,
          circularRefs
        },
        typeStats
      });
      (0, _dumpHelpers.replaceWithCode)(code, callExpr);
    }
  }

  const detailExportName = "createDetailedValidator";

  if (references[detailExportName]) {
    for (const path of references[detailExportName]) {
      const callExpr = path.parentPath;
      const [finalIR, typeStats] = finalizeType(path, instantiatedTypes, namedTypes);
      const {
        circularRefs,
        expectedValueAsIR
      } = (0, _macroAssertions.getOptions)("detailed", path, detailExportName);
      const code = (0, _irToInline.default)(finalIR, {
        instantiatedTypes,
        options: {
          errorMessages: true,
          circularRefs,
          expectedValueAsIR
        },
        typeStats
      });
      (0, _dumpHelpers.replaceWithCode)(code, callExpr);
    }
  }
}

var _default = (0, _babelPluginMacros.createMacro)(macroHandler);

exports.default = _default;